#include <boost/algorithm/string/trim.hpp>

#include "localization.h"
#include "query.h"

#include <unity/scopes/Annotation.h>
#include <unity/scopes/CategorisedResult.h>
#include <unity/scopes/CategoryRenderer.h>
#include <unity/scopes/QueryBase.h>
#include <unity/scopes/SearchReply.h>
#include <unity/scopes/OptionSelectorFilter.h>
#include <unity/scopes/FilterOption.h>
#include <unity/scopes/RangeInputFilter.h>
#include <unity/scopes/ValueSliderFilter.h>
#include <unity/scopes/ValueSliderLabels.h>
#include <unity/scopes/VariantBuilder.h>

#include "books.h"

#include <set>
#include <sstream>
#include <iomanip>
#include <thread>
#include <chrono>
#include <cmath>

namespace sc = unity::scopes;

using namespace std;
using namespace scope;

/**
 * Define the layout for theresults
 *
 * The icon size is medium, and ask for the card layout
 * itself to be horizontal. I.e. the text will be placed
 * next to the image.
 */
const static string CATEGORY_TEMPLATE =
        R"(
{
        "schema-version": 1,
        "template": {
            "category-layout": "grid",
            "card-layout": "horizontal",
            "card-size": "medium"
        },
        "components": {
            "title": "title",
            "art" : {
                "field": "art"
            },
            "subtitle": "subtitle",
            "attributes": {
                "field": "attributes",
                "max-count": 2
            }
        }
    }
)";

Query::Query(const sc::CannedQuery &query, const sc::SearchMetadata &metadata, const std::string& scopedir) :
    sc::SearchQueryBase(query, metadata),
    scope_dir_(scopedir) {
}

void Query::cancelled() {
}

bool match(Book const& book,
        std::string const& query,
        std::string const& department,
        std::string const& format,
        std::set<std::string> const& categories,
        double min_price,
        double max_price,
        double rating)
{
    if (department.size() && book.department != department) {
        return false;
    }

    if (format.size() && book.format != format) {
        return false;
    }

    if (categories.size() && categories.find(book.category) == categories.end()) {
        return false;
    }

    if (book.price < min_price || book.price > max_price) {
        return false;
    }

    if (book.rating < rating) {
        return false;
    }

    if (query.size()) {
        if ((book.title.find(query) == std::string::npos) && (book.author.find(query) == std::string::npos)) {
            return false;
        }
    }

    return true;
}

void Query::generate_results(sc::SearchReplyProxy const& reply,
        std::string const& book_format,
        std::set<std::string> const& book_categories,
        double min_price,
        double max_price,
        double rating)
{
    auto const query_string = query().query_string();
    auto const dept_id = query().department_id();

    const bool show_single_cat = (dept_id.size() > 0);

    rating = std::round(rating);

    int count = 0;
    if (show_single_cat) {
        for (auto const& book: books) {
            if (match(book, query_string, dept_id, book_format, book_categories, min_price, max_price, rating)) {
                ++count;
            }
        }
    }

    sc::Category::SCPtr cat_single = reply->register_category("all", std::to_string(count) + " Books", "", sc::CategoryRenderer(CATEGORY_TEMPLATE));
    sc::Category::SCPtr cat_fiction = reply->register_category("fiction", "Fiction", "", sc::CategoryRenderer(CATEGORY_TEMPLATE));
    sc::Category::SCPtr cat_children = reply->register_category("childrens", "Children", "", sc::CategoryRenderer(CATEGORY_TEMPLATE));

    for (auto const& book: books)
    {
        if (match(book, query_string, dept_id, book_format, book_categories, min_price, max_price, rating)) {
            std::this_thread::sleep_for(std::chrono::milliseconds(50));

            sc::Category::SCPtr cat;
            if (show_single_cat) {
                cat = cat_single;
            } else {
                if (book.department == "fiction") {
                    cat = cat_fiction;
                } else {
                    cat = cat_children;
                }
            }
            sc::CategorisedResult res(cat);
            res.set_uri("uri" + std::to_string(count++));
            res.set_title(book.title);
            res.set_art(scope_dir_ + "/covers/" + book.department + "/" + book.art);
            res["subtitle"] = book.author;
            res["description"] = book.synopsis;

            std::stringstream rating;
            rating << std::fixed << std::setprecision(1);
            rating << "☆ " << book.rating;

            std::stringstream price;
            price << std::fixed << std::setprecision(2);
            price << "£" << book.price;

            res["price"] = price.str();
            res["rating"] = rating.str();

            sc::VariantBuilder builder;
            builder.add_tuple({
                {"value", sc::Variant(price.str())},
            });

            builder.add_tuple({
                    {"value", sc::Variant(rating.str())},
            });
            res["attributes"] = builder.end();

            // Push the result
            if (!reply->push(res)) {
                return;
            }
        }
    }
}

void Query::run(sc::SearchReplyProxy const& reply) {
    auto const query_string = query().query_string();
    auto const dept_id = query().department_id();
    auto filter_state = query().filter_state();

    try {
        // departments
        {
            sc::Department::SPtr root = sc::Department::create("", query(), "All Books");
            sc::Department::SPtr tvs = sc::Department::create("fiction", query(), "Fiction");
            sc::Department::SPtr pcs = sc::Department::create("childrens", query(), "Children's");
            root->set_subdepartments({tvs, pcs});

            reply->register_departments(root);
        }

        sc::RangeInputFilter::SPtr price_filter = sc::RangeInputFilter::create("price", "Min", "", "to", "Max", "");
        price_filter->set_title("Price range");

        sc::ValueSliderFilter::SPtr rating_filter = sc::ValueSliderFilter::create("rating", 1.0f, 5.0f, 1.0f, sc::ValueSliderLabels("1", "5", {{2.0f, "2"}, {3.0f, "3"}, {4.0f, "4"}}));
        rating_filter->set_title("Rating");

        sc::OptionSelectorFilter::SPtr category_filter = sc::OptionSelectorFilter::create("bookcategory", "Top picks", true);
        category_filter->add_option("book of the month", "Book of the month");
        category_filter->add_option("bestseller", "Bestsellers");
        category_filter->add_option("special edition", "Special editions");
        category_filter->add_option("recommended", "Recommended");

        sc::OptionSelectorFilter::SPtr format_filter = sc::OptionSelectorFilter::create("format", "Format", false);
        format_filter->add_option("hardback", "Hardcover");
        format_filter->add_option("paperback", "Paperback");
        format_filter->add_option("audiobook", "Audiobook");

        std::string book_format;
        std::set<std::string> book_categories;
        double min_price = 0.0f;
        double max_price = 9999.0f;
        double rating = 0.0f;

        reply->push({category_filter, format_filter, price_filter, rating_filter}, query().filter_state());

        // determine filter values
        auto format = format_filter->active_options(filter_state);
        if (format.size()) {
            book_format = (*format.begin())->id();
        }

        if (price_filter->has_start_value(filter_state)) {
            min_price = price_filter->start_value(filter_state);
        }

        if (price_filter->has_end_value(filter_state)) {
            max_price = price_filter->end_value(filter_state);
        }

        if (rating_filter->has_value(filter_state)) {
            rating = rating_filter->value(filter_state);
        }

        if (category_filter->has_active_option(filter_state)) {
            auto cats = category_filter->active_options(filter_state);
            for (auto const &opt: cats) {
                book_categories.insert(opt->id());
            }
        }
        generate_results(reply, book_format, book_categories, min_price, max_price, rating);

    } catch (std::exception &e) {
        cerr << e.what() << endl;
        reply->error(current_exception());
    }
}
